@@ -146,7 +146,7 @@ struct client {
     size_t nwrite;
   } stream;
 
-  ngtcp2_connection_close_error last_error;
+  ngtcp2_ccerr last_error;
 
   ev_io rev;
   ev_timer timer;
@@ -443,11 +443,10 @@ static int client_read(struct client *c) {
       fprintf(stderr, "ngtcp2_conn_read_pkt: %s\n", ngtcp2_strerror(rv));
       if (!c->last_error.error_code) {
         if (rv == NGTCP2_ERR_CRYPTO) {
-          ngtcp2_connection_close_error_set_transport_error_tls_alert(
+          ngtcp2_ccerr_set_tls_alert(
               &c->last_error, ngtcp2_conn_get_tls_alert(c->conn), NULL, 0);
         } else {
-          ngtcp2_connection_close_error_set_transport_error_liberr(
-              &c->last_error, rv, NULL, 0);
+          ngtcp2_ccerr_set_liberr(&c->last_error, rv, NULL, 0);
         }
       }
       return -1;
@@ -506,7 +505,7 @@ static int client_write_streams(struct client *c) {
   ngtcp2_tstamp ts = timestamp();
   ngtcp2_pkt_info pi;
   ngtcp2_ssize nwrite;
-  uint8_t buf[1280];
+  uint8_t buf[1452];
   ngtcp2_path_storage ps;
   ngtcp2_vec datav;
   size_t datavcnt;
@@ -536,8 +535,7 @@ static int client_write_streams(struct client *c) {
       default:
         fprintf(stderr, "ngtcp2_conn_writev_stream: %s\n",
                 ngtcp2_strerror((int)nwrite));
-        ngtcp2_connection_close_error_set_transport_error_liberr(
-            &c->last_error, (int)nwrite, NULL, 0);
+        ngtcp2_ccerr_set_liberr(&c->last_error, (int)nwrite, NULL, 0);
         return -1;
       }
     }
@@ -593,8 +591,8 @@ static void client_close(struct client *c) {
   ngtcp2_path_storage ps;
   uint8_t buf[1280];
 
-  if (ngtcp2_conn_is_in_closing_period(c->conn) ||
-      ngtcp2_conn_is_in_draining_period(c->conn)) {
+  if (ngtcp2_conn_in_closing_period(c->conn) ||
+      ngtcp2_conn_in_draining_period(c->conn)) {
     goto fin;
   }
 
@@ -655,7 +653,7 @@ static int client_init(struct client *c) {
 
   memset(c, 0, sizeof(*c));
 
-  ngtcp2_connection_close_error_default(&c->last_error);
+  ngtcp2_ccerr_default(&c->last_error);
 
   c->fd = create_sock((struct sockaddr *)&remote_addr, &remote_addrlen,
                       REMOTE_HOST, REMOTE_PORT);
