@@ -24,6 +24,7 @@
  */
 #include "tls_client_context_boringssl.h"
 
+#include <cstring>
 #include <iostream>
 #include <fstream>
 
@@ -33,6 +34,7 @@
 
 #include "client_base.h"
 #include "template.h"
+#include "tls_shared_boringssl.h"
 
 extern Config config;
 
@@ -48,6 +50,11 @@ SSL_CTX *TLSClientContext::get_native_handle() const { return ssl_ctx_; }
 
 namespace {
 int new_session_cb(SSL *ssl, SSL_SESSION *session) {
+  auto conn_ref = static_cast<ngtcp2_crypto_conn_ref *>(SSL_get_app_data(ssl));
+  auto c = static_cast<ClientBase *>(conn_ref->user_data);
+
+  c->ticket_received();
+
   auto f = BIO_new_file(config.session_file, "w");
   if (f == nullptr) {
     std::cerr << "Could not write TLS session in " << config.session_file
@@ -82,8 +89,8 @@ int TLSClientContext::init(const char *private_key_file,
 
   SSL_CTX_set_default_verify_paths(ssl_ctx_);
 
-  if (SSL_CTX_set1_curves_list(ssl_ctx_, config.groups) != 1) {
-    std::cerr << "SSL_CTX_set1_curves_list failed" << std::endl;
+  if (SSL_CTX_set1_groups_list(ssl_ctx_, config.groups) != 1) {
+    std::cerr << "SSL_CTX_set1_groups_list failed" << std::endl;
     return -1;
   }
 
@@ -108,6 +115,15 @@ int TLSClientContext::init(const char *private_key_file,
     SSL_CTX_sess_set_new_cb(ssl_ctx_, new_session_cb);
   }
 
+#ifdef HAVE_LIBBROTLI
+  if (!SSL_CTX_add_cert_compression_alg(
+          ssl_ctx_, ngtcp2::tls::CERTIFICATE_COMPRESSION_ALGO_BROTLI,
+          ngtcp2::tls::cert_compress, ngtcp2::tls::cert_decompress)) {
+    std::cerr << "SSL_CTX_add_cert_compression_alg failed" << std::endl;
+    return -1;
+  }
+#endif // HAVE_LIBBROTLI
+
   return 0;
 }
 
