@@ -29,7 +29,13 @@
 #  include <config.h>
 #endif /* HAVE_CONFIG_H */
 
-void test_ngtcp2_ringbuf_push_front(void);
-void test_ngtcp2_ringbuf_pop_front(void);
+#define MUNIT_ENABLE_ASSERT_ALIASES
+
+#include "munit.h"
+
+extern const MunitSuite ringbuf_suite;
+
+munit_void_test_decl(test_ngtcp2_ringbuf_push_front);
+munit_void_test_decl(test_ngtcp2_ringbuf_pop_front);
 
 #endif /* NGTCP2_RINGBUF_TEST_H */
